/*
 * Copyright Bigeon Emmanuel (2014)
 *
 * emmanuel@bigeon.fr
 *
 * This software is a computer program whose purpose is to
 * provide a generic framework for console applications.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */
/**
 * gclc:fr.bigeon.gclc.command.ParametrizedCommandTest.java
 * Created on: Nov 18, 2016
 */
package fr.bigeon.gclc.command;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.io.PrintStream;
import java.nio.charset.StandardCharsets;

import org.junit.Test;

import fr.bigeon.gclc.exception.CommandRunException;
import fr.bigeon.gclc.exception.InvalidParameterException;
import fr.bigeon.gclc.manager.ConsoleInput;
import fr.bigeon.gclc.manager.ConsoleOutput;
import fr.bigeon.gclc.manager.PipedConsoleInput;
import fr.bigeon.gclc.manager.PipedConsoleOutput;

/**
 * <p>
 * TODO
 *
 * @author Emmanuel Bigeon */
public class ParametrizedCommandTest {

    /** Test method for
     * {@link fr.bigeon.gclc.command.ParametrizedCommand#addParameter(java.lang.String, boolean, boolean)}.
     *
     * @throws InvalidParameterException */
    @Test
    public final void testAddParameter() throws InvalidParameterException {
        ParametrizedCommand cmd = new ParametrizedCommand("name") {

            @Override
            protected void doExecute(final ConsoleOutput out,
                                     final ConsoleInput in,
                                     final CommandParameters parameters) {
                //
            }

            @Override
            public String tip() {
                return null;
            }

            @Override
            protected String usageDetail() {
                return null;
            }
        };
        cmd = new ParametrizedCommand("name", true) {

            @Override
            protected void doExecute(final ConsoleOutput out,
                                     final ConsoleInput in,
                                     final CommandParameters parameters) {
                //
            }

            @Override
            public String tip() {
                return null;
            }

            @Override
            protected String usageDetail() {
                return null;
            }
        };
        // XXX Boolean flag should not be specified mandatory! They are by
        // nature qualified
        final String str = "str";
        assertTrue(cmd.getBooleanParameters().isEmpty());
        assertTrue(cmd.getStringParameters().isEmpty());
        cmd.addBooleanParameter("boolFlag");
        assertEquals(1, cmd.getBooleanParameters().size());
        assertTrue(cmd.getStringParameters().isEmpty());
        cmd.addStringParameter(str, false);
        assertEquals(1, cmd.getBooleanParameters().size());
        assertEquals(1, cmd.getStringParameters().size());
        assertFalse(cmd.isNeeded(str));
        cmd.addBooleanParameter("boolFlag");
        assertEquals(1, cmd.getBooleanParameters().size());
        assertEquals(1, cmd.getStringParameters().size());
        cmd.addStringParameter(str, true);
        assertEquals(1, cmd.getBooleanParameters().size());
        assertEquals(1, cmd.getStringParameters().size());
        assertTrue(cmd.isNeeded(str));
        cmd.addStringParameter(str, false);
        assertEquals(1, cmd.getBooleanParameters().size());
        assertEquals(1, cmd.getStringParameters().size());
        assertTrue(cmd.isNeeded(str));
    }

    /** Test method for
     * {@link fr.bigeon.gclc.command.ParametrizedCommand#execute(java.lang.String[])}.
     *
     * @throws CommandRunException
     * @throws InterruptedException
     * @throws IOException */
    @Test
    public final void testExecute() throws CommandRunException,
                                    InterruptedException, IOException {
        final String addParam = "additional";
        final String str1 = "str1";
        final String str2 = "str2";
        final String bool1 = "bool1";
        final String bool2 = "bool2";

        // Test on command with no needed elements
        ParametrizedCommand cmd = new ParametrizedCommand("name", false) {
            private boolean evenCall = true;

            @Override
            protected void doExecute(final ConsoleOutput out,
                                     final ConsoleInput in,
                                     final CommandParameters parameters) {
                assertTrue(parameters.getBooleanArgumentKeys().isEmpty());
                assertTrue(parameters.getStringArgumentKeys().isEmpty());
                if (evenCall) {
                    assertTrue(parameters.getAdditionals().isEmpty());
                    evenCall = false;
                } else {
                    assertEquals(1, parameters.getAdditionals().size());
                    assertTrue(parameters.getAdditionals().contains(addParam));
                    evenCall = true;
                }
            }

            @Override
            public String tip() {
                return "";
            }

            @Override
            protected String usageDetail() {
                return null;
            }
        };
        cmd.execute(null, null);
        cmd.execute(null, null, "-" + addParam);
        cmd.execute(null, null, addParam);
        cmd.execute(null, null, "-" + addParam, addParam);
        cmd = new ParametrizedCommand("name", false) {
            private int call = 0;
            {

                try {
                    addStringParameter(str1, false);
                    addStringParameter(str2, false);
                    addBooleanParameter(bool1);
                    addBooleanParameter(bool2);
                } catch (final InvalidParameterException e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                }
            }

            @Override
            protected void doExecute(final ConsoleOutput out,
                                     final ConsoleInput in,
                                     final CommandParameters parameters) {
                assertEquals(2, parameters.getBooleanArgumentKeys().size());
                assertEquals(2, parameters.getStringArgumentKeys().size());
                switch (call) {
                case 0:
                case 1:
                case 2:
                case 3:
                    assertNull(parameters.get(str1));
                    assertNull(parameters.get(str2));
                    assertFalse(parameters.getBool(bool1));
                    assertFalse(parameters.getBool(bool2));
                    call++;
                    break;
                case 4:
                    assertEquals(str2, parameters.get(str1));
                    assertNull(parameters.get(str2));
                    assertFalse(parameters.getBool(bool1));
                    assertFalse(parameters.getBool(bool2));
                    call++;
                    break;
                case 5:
                    assertEquals(str2, parameters.get(str1));
                    assertNull(parameters.get(str2));
                    assertTrue(parameters.getBool(bool1));
                    assertFalse(parameters.getBool(bool2));
                    call = 0;
                    break;
                default:
                    break;
                }
            }

            @Override
            public String tip() {
                return "";
            }

            @Override
            protected String usageDetail() {
                return null;
            }
        };
        cmd.execute(null, null);
        cmd.execute(null, null, "-" + addParam);
        cmd.execute(null, null, addParam);
        cmd.execute(null, null, "-" + addParam, addParam);
        cmd.execute(null, null, "-" + str1, str2);
        cmd.execute(null, null, "-" + str1, str2, "-" + bool1);
        cmd = new ParametrizedCommand("name", true) {
            private int call = 0;
            {

                try {
                    addStringParameter(str1, false);
                    addStringParameter(str2, false);
                    addBooleanParameter(bool1);
                    addBooleanParameter(bool2);
                } catch (final InvalidParameterException e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                }
            }

            @Override
            protected void doExecute(final ConsoleOutput out,
                                     final ConsoleInput in,
                                     final CommandParameters parameters) {
                assertEquals(2, parameters.getBooleanArgumentKeys().size());
                assertEquals(2, parameters.getStringArgumentKeys().size());
                switch (call) {
                case 0:
                    assertNull(parameters.get(str1));
                    assertNull(parameters.get(str2));
                    assertFalse(parameters.getBool(bool1));
                    assertFalse(parameters.getBool(bool2));
                    call++;
                    break;
                case 1:
                    assertEquals(str2, parameters.get(str1));
                    assertNull(parameters.get(str2));
                    assertFalse(parameters.getBool(bool1));
                    assertFalse(parameters.getBool(bool2));
                    call++;
                    break;
                case 2:
                    assertEquals(str2, parameters.get(str1));
                    assertNull(parameters.get(str2));
                    assertTrue(parameters.getBool(bool1));
                    assertFalse(parameters.getBool(bool2));
                    call = 0;
                    break;
                default:
                    break;
                }
            }

            @Override
            public String tip() {
                return "";
            }

            @Override
            protected String usageDetail() {
                return null;
            }
        };
        cmd.execute(null, null);
        cmd.execute(null, null, "-" + str1, str2);
        cmd.execute(null, null, "-" + str1, str2, "-" + bool1);
        try {
            cmd.execute(null, null, addParam);
            fail("Strict should fail with unexpected argument");
        } catch (final CommandRunException e) {
            // ok
        }
        try {
            cmd.execute(null, null, "-" + addParam);
            fail("Strict should fail with unexpected argument");
        } catch (final CommandRunException e) {
            // ok
        }
        // Test on command with missing needed elements
        cmd = new ParametrizedCommand("name", false) {
            private final int call = 0;
            {

                try {
                    addStringParameter(str1, true);
                    addStringParameter(str2, false);
                    addBooleanParameter(bool1);
                    addBooleanParameter(bool2);
                } catch (final InvalidParameterException e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                }
            }

            @Override
            protected void doExecute(final ConsoleOutput out,
                                     final ConsoleInput in,
                                     final CommandParameters parameters) {
                assertEquals(str2, parameters.get(str1));
            }

            @Override
            public String tip() {
                return "";
            }

            @Override
            protected String usageDetail() {
                return null;
            }
        };
        cmd.execute(null, null, "-" + str1, str2);
        cmd.execute(null, null, "-" + str1, str2, "-" + bool1);
        cmd.execute(null, null, "-" + str1, str2, "-" + addParam);
        cmd.execute(null, null, "-" + str1, str2, addParam);
        cmd.execute(null, null, "-" + str1, str2, "-" + addParam, addParam);
        try {
            cmd.execute(null, null);
            fail("needed " + str1 + " not provided shall fail");
        } catch (final CommandRunException e) {
            // ok
        }
        cmd = new ParametrizedCommand("name", true) {
            private final int call = 0;
            {

                try {
                    addStringParameter(str1, true);
                    addStringParameter(str2, false);
                    addBooleanParameter(bool1);
                    addBooleanParameter(bool2);
                } catch (final InvalidParameterException e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                }
            }

            @Override
            protected void doExecute(final ConsoleOutput out,
                                     final ConsoleInput in,
                                     final CommandParameters parameters) {
                //
                assertEquals(str2, parameters.get(str1));
            }

            @Override
            public String tip() {
                return "";
            }

            @Override
            protected String usageDetail() {
                return null;
            }
        };
        cmd.execute(null, null, "-" + str1, str2);
        cmd.execute(null, null, "-" + str1, str2, "-" + bool1);
        try {
            cmd.execute(null, null, "-" + str1, str2, addParam);
            fail("Additional parameter should cause failure");
        } catch (final CommandRunException e) {
            // ok
        }
        try {
            cmd.execute(null, null);
            fail("needed " + str1 + " not provided shall fail");
        } catch (final CommandRunException e) {
            // ok
        }
        try {
            cmd.execute(null, null, "-" + str1, str2, "-" + addParam);
            fail("unepected error");
        } catch (final CommandRunException e) {
            // ok
        }
        try {
            cmd.execute(null, null, "-" + str1, str2, "-" + addParam, addParam);
            fail("unepected error");
        } catch (final CommandRunException e) {
            // ok
        }
        // TODO Test of interactive not providing and providing all needed
    }

    @Test
    public void testExecuteInteractive() throws IOException,
                                         CommandRunException,
                                         InterruptedException {
        ParametrizedCommand cmd;
        final String addParam = "additional";
        final String str1 = "str1";
        final String str2 = "str2";
        final String bool1 = "bool1";
        final String bool2 = "bool2";

        cmd = new ParametrizedCommand("name", false) {
            {
                try {
                    addStringParameter(str1, true);
                    addStringParameter(str2, false);
                    addBooleanParameter(bool1);
                    addBooleanParameter(bool2);
                } catch (final InvalidParameterException e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                }
            }

            @Override
            protected void doExecute(final ConsoleOutput out,
                                     final ConsoleInput in,
                                     final CommandParameters parameters) throws CommandRunException {
                if (!str2.equals(parameters.get(str1))) {
                    throw new CommandRunException("Expected other argument",
                            this);
                }
            }

            @Override
            public String tip() {
                return "";
            }

            @Override
            protected String usageDetail() {
                return null;
            }
        };
        try (PipedConsoleOutput out = new PipedConsoleOutput();
             PipedOutputStream pout = new PipedOutputStream();
             PipedInputStream pis = new PipedInputStream(pout);
             BufferedReader buf = new BufferedReader(
                     new InputStreamReader(pis, StandardCharsets.UTF_8));
             PipedConsoleInput in = new PipedConsoleInput(
                     new PrintStream(pout))) {
            cmd.execute(out, in, "-" + str1, str2);
            cmd.execute(out, in, "-" + str1, str2, "-" + bool1);
            cmd.execute(out, in, "-" + str1, str2, addParam);
            cmd.execute(out, in, "-" + str1, str2, "-" + addParam);
            cmd.execute(out, in, "-" + str1, str2, "-" + addParam, addParam);

            final Thread th = new Thread(new Runnable() {

                @Override
                public void run() {
                    try {
                        assertEquals("value of " + str1 + "? ", buf.readLine());
                        in.type("");
                        assertEquals(
                                "value of " + str1 + "? (cannot be empty) ",
                                buf.readLine());
                        in.type("");
                        assertEquals(
                                "value of " + str1 + "? (cannot be empty) ",
                                buf.readLine());
                        in.type(str2);
                    } catch (final IOException e) {
                        assertNull(e);
                    }
                }
            });
            th.start();

            cmd.execute(out, in);

            th.join();
        }
        try (PipedConsoleOutput out = new PipedConsoleOutput();
             PipedOutputStream pout = new PipedOutputStream();
             PipedInputStream pis = new PipedInputStream(pout);
             BufferedReader buf = new BufferedReader(
                     new InputStreamReader(pis, StandardCharsets.UTF_8));
             PipedConsoleInput in = new PipedConsoleInput(
                     new PrintStream(pout))) {

            final Thread th = new Thread(new Runnable() {

                @Override
                public void run() {
                    try {
                        assertEquals("value of " + str1 + "? ", buf.readLine());
                        in.type(str2);
                    } catch (final IOException e) {
                        assertNull(e);
                    }
                }
            });
            th.start();

            cmd.execute(out, in, "-" + addParam);

            th.join();
        }
        try {
            final PipedConsoleOutput out = new PipedConsoleOutput();
            final PipedConsoleInput test = new PipedConsoleInput(null);
            test.close();
            out.close();
            cmd.execute(out, test, "-" + str1, str2);
            cmd.execute(out, test, "-" + addParam);
            fail("Closed manager shall cause error");
        } catch (final CommandRunException e) {
            // ok
        }
    }

    /** Test method for
     * {@link fr.bigeon.gclc.command.ParametrizedCommand#ParametrizedCommand(fr.bigeon.gclc.manager.ConsoleManager, java.lang.String)}. */
    @Test
    public final void testParametrizedCommandConsoleManagerString() {
        ParametrizedCommand cmd = new ParametrizedCommand("name") {

            @Override
            protected void doExecute(final ConsoleOutput out,
                                     final ConsoleInput in,
                                     final CommandParameters parameters) {
                //
            }

            @Override
            public String tip() {
                return null;
            }

            @Override
            protected String usageDetail() {
                return null;
            }
        };
        assertTrue(cmd.isStrict());
        cmd = new ParametrizedCommand("name") {
            @Override
            protected void doExecute(final ConsoleOutput out,
                                     final ConsoleInput in,
                                     final CommandParameters parameters) {
                //
            }

            @Override
            public String tip() {
                return null;
            }

            @Override
            protected String usageDetail() {
                return null;
            }
        };
        assertTrue(cmd.isStrict());
        cmd = new ParametrizedCommand("name") {
            @Override
            protected void doExecute(final ConsoleOutput out,
                                     final ConsoleInput in,
                                     final CommandParameters parameters) {
                //
            }

            @Override
            public String tip() {
                return null;
            }

            @Override
            protected String usageDetail() {
                return null;
            }
        };
        assertTrue(cmd.isStrict());
    }

    /** Test method for
     * {@link fr.bigeon.gclc.command.ParametrizedCommand#ParametrizedCommand(fr.bigeon.gclc.manager.ConsoleManager, java.lang.String, boolean)}. */
    @Test
    public final void testParametrizedCommandConsoleManagerStringBoolean() {
        ParametrizedCommand cmd = new ParametrizedCommand("name", false) {
            @Override
            protected void doExecute(final ConsoleOutput out,
                                     final ConsoleInput in,
                                     final CommandParameters parameters) {
                //
            }

            @Override
            public String tip() {
                return null;
            }

            @Override
            protected String usageDetail() {
                return null;
            }
        };
        assertFalse(cmd.isStrict());
        cmd = new ParametrizedCommand("name", false) {
            @Override
            protected void doExecute(final ConsoleOutput out,
                                     final ConsoleInput in,
                                     final CommandParameters parameters) {
                //
            }

            @Override
            public String tip() {
                return null;
            }

            @Override
            protected String usageDetail() {
                return null;
            }
        };
        assertFalse(cmd.isStrict());
        cmd = new ParametrizedCommand("name", false) {
            @Override
            protected void doExecute(final ConsoleOutput out,
                                     final ConsoleInput in,
                                     final CommandParameters parameters) {
                //
            }

            @Override
            public String tip() {
                return null;
            }

            @Override
            protected String usageDetail() {
                return null;
            }
        };
        assertFalse(cmd.isStrict());
    }

}
